
/* main.c - Thread calculation and mutual exclusion */
/* Written by Ryan O'Neill */

#include "abyss.h"

char errbuf[BUF_LEN];
pcap_t *p_handler;
int linktype;
int global_thread_flag = 0;
char *dev;

struct hostent *hptr2;
struct options opts;
unsigned long local_ip;

char *subopts[] = {
#define BANNER 0
  "banner",
#define PORT   1
  "port",
  NULL
};

#define OS_PRINTS "os.prints"

/* Our calls to _scan_() and tcppackets() must be locked */

pthread_mutex_t scanlock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t pcaplock = PTHREAD_MUTEX_INITIALIZER;

/* We are using these to keep track of threads, and properly */
/* employing new ones when others finish, mutual exclusion is significant here */

pthread_mutex_t finlock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t fincond = PTHREAD_COND_INITIALIZER;
int maxconn = 0;
int finished = 0;

pthread_t passive_thread;

void usage (void);

struct host host[MAX_THREAD];

/* This is actually a dummy function, and _scan_() is the real one */
/* This is used so that after calling _scan_() it can notify the calling thread */
/* That it is finished, so that another thread can quickly takes its place with */
/* The next host. 3 threads is default number of threads scanning at the same time */

void
catch_sigint (int sig)
{
  Logprint ("Caught Signal CTRL-C ! AbySS Exiting\n");
  pthread_exit (passive_thread);
}

void *
scan (void *arg)
{
  struct host *hinfo = (struct host *) arg;
  unsigned long ipaddr;
  char *host_name;

  if (!opts.singlehost)
    {
      ipaddr = hinfo->ip;
      char *host_name = hinfo->name;
    }
  else if (opts.singlehost)
    {
      ipaddr = ntohl (hinfo->ip);
    }
  /* _scan_() is the actual scanning function */

  if (opts.fromlist)
    {
      char *host_name = hinfo->name;
      if ((ipaddr = resolve_ip (host_name, hptr2)) == -1)
	{
	  Logprint ("AbysS [Net] - Could not resolve %s\n", host_name);
	  goto thread_done;
	}
    }

  _scan_ (ipaddr);
  goto thread_done;


  /* Here we go into the lock, and after _scan_() is done create        */
  /* The thread condition signal for finished++ notification in main()  */
  /* Basically this sets up a condition signal, so when finished++      */
  /* increments, certain flags are set, notifying the thread/socket     */
  /* pair are finished, and pthread_cond_signal sends a signal to the   */
  /* Calling thread main() saying this thread is finished, create a new */
  /* thread to call scan()                                              */

thread_done:
  pthread_mutex_lock (&finlock);
  finished++;
  if (opts.singlehost)
    global_thread_flag |= DONE;
  else
    hinfo->socket.flag |= DONE;
  pthread_cond_signal (&fincond);
  pthread_mutex_unlock (&finlock);
}

int
main (int argc, char **argv)
{

  struct hostent *hptr;
  extern char *optarg;
  char *options, *value, *p, *temp, **ptr, *p2, *p3, *pc;
  char *paddr, str[MAX_STRING];
  char buf[MAX_STRING];
  FILE *osfd;

  struct os_list *osptr = oslist;

  /* time */
  time_t clock;
  struct tm *tm;
  struct servent *spt;
  char time_buf[BUF_LEN];
  char OSbuf[MAX_STRING];

  unsigned long int net;
  FILE *fd;

  int i, j, k, c, err, serv;
  int filter_len = 0;
  int ret = 0;
  int nconn, maxnconn;

  struct icmp_ip
  {
    u_long start;
    u_long end;
  } icmp_ip;

  pthread_t scantid[MAX_THREAD];
  pthread_t sniff_thread;
  pthread_t tid;
  pthread_t syn_sniff_thread;
  pthread_t icmp_sniff_thread;

  char *filter[10];
  char *filtr;
  char *tmp;

  struct sigaction act;
  sigset_t set;

  act.sa_handler = catch_sigint;
  sigemptyset (&act.sa_mask);
  act.sa_flags = 0;
  sigaction (SIGINT, &act, NULL);
  sigemptyset (&set);
  sigaddset (&set, SIGINT);

  if (argc < 3)
    {
      print_banner ();
      usage ();
    }

  opts.threads = NUM_THREAD;
  opts.fromlist = 0;
  opts.singlehost = 0;
  opts.logging = 0;
  opts.banner = 0;
  opts.pingnet = 0;
  opts.subscan = 0;
  opts.verbose = 0;
  opts.interface = 0;
  opts.synscan = 0;
  opts.passivemode = 0;
  opts.payload = 0;
  opts.headers = 0;
  opts.bpf = 0;

  i = 0;
  err = 0;

  if (getuid () != 0)
    {
      printf ("AbysS [Error] - Must be root to run AbysS\n");
      exit (0);
    }


  while ((c = getopt (argc, argv, "hblvs:f:n:O:p:o:t:i:S:b:")) != -1)
    {
      switch (c)
	{
	case 's':
	  opts.singlehost = 1;
	  opts.host = (char *) malloc (strlen (optarg) + 1);
	  strcpy (opts.host, optarg);
	  break;
	case 'f':
	  opts.fromlist = 1;
	  opts.hostlist = strdup (optarg);
	  break;
	case 'n':
	  opts.subscan = 1;
	  opts.host = malloc (strlen (optarg) + 1);
	  strcpy (opts.host, optarg);
	  printf ("Netmask host %s\n", opts.host);
	  break;
	case 'p':
	  opts.pingnet = 1;
	  opts.host = malloc (strlen (optarg + 1));
          strcpy(opts.host, optarg);
	
          if (!argv[3])
            {
              printf("-p requires one more argument [local ip]\n");
              exit(0);
            }
          
          opts.local_net = (char *)malloc(strlen(argv[3]) + 1);
          strcpy(opts.local_net, argv[3]);
          
          if ((opts.myip = inet_addr(opts.local_net)) == -1)
            {
              printf("Your local ip must be in dotted decimal form\n");
              exit(-1);
            }

          break;
	case 'o':
	  opts.logging = 1;
	  opts.logfile = malloc (strlen (optarg) + 1);
	  strcpy (opts.logfile, optarg);
	  if ((opts.logfd = fopen (opts.logfile, "w")) == NULL)
	    fprintf (stderr, "AbysS - [Error] Unable to log\n");
	  break;
	case 't':
	  opts.threads = atoi (optarg);
	  break;
	case 'v':
	  opts.verbose = 1;
	  break;
	case 'h':
	  opts.headers = 1;
	  break;
	case 'l':
	  opts.payload = 1;
	  break;
	case 'i':
	  opts.interface = 1;
	  dev = malloc (strlen (optarg) + 1);
	  strcpy (dev, optarg);
	  break;
	case 'S':
	  opts.synscan = 1;
	  opts.host = malloc (strlen (optarg) + 1);
	  strcpy (opts.host, optarg);
	  break;
	case 'O':
	  opts.passivemode = 1;
	  dev = malloc (strlen (optarg) + 1);
	  strncpy (dev, optarg, strlen (optarg) - 1);

	  if (argv[3])
	    if (strcmp (argv[3], "-b") == 0)
	      {
		opts.bpf = 1;
		for (i = 0; i < argc - 1; i++)
		  {
		    if ((filter[i] = strdup (argv[i + 1])) == NULL)
		      {
			printf ("Could not allocate memory!\n");
			exit (-1);
		      }
		    filter_len += strlen (filter[i]);
		  }

		++filter_len;

		if ((filtr = (char *) malloc (filter_len)) == NULL)
		  {
		    printf ("Could not allocate memory!\n");
		    exit (-1);
		  }
		if ((tmp = (char *) malloc (filter_len)) == NULL)
		  {
		    printf ("Could not allocate memory!\n");
		    exit (-1);
		  }
		k = 0;
		for (i = 2; i < argc - 2; i++)
		  {
		    c = strlen (filter[i + 1]);
		    strcpy (tmp, filter[i + 1]);
		    for (j = 0; j < c; j++)
		      filtr[k++] = tmp[j];
		    filtr[k++] = ' ';
		    memset (tmp, 0, sizeof (tmp));
		  }
		filtr[k] = '\0';

		if (strstr (filtr, "-h"))
		  {
		    opts.headers = 1;
		    if ((pc = strchr (filtr, '-')) != NULL)
		      {
			*pc = '\0';
			*pc++ = '\0';
		      }
		  }
	      }
	  break;
	case 'B':
	  /* Process sub option banner */
	  options = optarg;

	  while (*options != '\0')
	    {

	      switch (getsubopt (&options, subopts, &value))
		{
		case BANNER:
		  if (value != NULL)
		    {
		      opts.banner = 1;
		      strncpy (banner_string.string, value, MAX_STRING);
		    }
		  break;
		case PORT:
		  if (opts.banner != 1)
		    {
		      printf
			("AbysS [Error] - Must assign banner=value first\n");
		      usage ();
		    }
		  banner_string.port = atoi (value);
		  break;
		default:
		  printf ("AbysS [Error] - Unknown option: %s \n", value);
		  usage ();
		  break;
		}
	    }
	  break;
	}
    }

  if ((serv = get_tcp_ports ("/etc/services")) < 1)
    {
      Logprint ("AbysS [Error] - Could not open /etc/services\n");
      exit (-1);
    }

  if ((opts.singlehost) && (opts.fromlist))
    {
      printf ("AbysS [Error] - Cannot use -s and -f together\n");
      exit (0);
    }

  if (opts.threads > MAX_THREAD)
    {
      printf ("AbysS [Error] - Too many threads - Max is 30\n");
      exit (0);
    }

  if ((osfd = fopen (OS_PRINTS, "r")) == NULL)
    {
      Logprint ("AbysS [Error] - Cannot open %s\n", OS_PRINTS);
      exit (-1);
    }

  if (opts.passivemode)
    {
      if (opts.bpf)
	{
	  if ((err = init_pcap_ ("%s", filtr)) < 0)
	    {
	      Logprint ("AbysS [Error] - Error opening up device %s\n", dev);
	      exit (-1);
	    }
	}
      else
	{
	  if ((err = init_pcap_ ("tcp")) < 0)
	    {
	      Logprint ("AbysS [Error] - Error opening device %s\n", dev);
	      exit (-1);
	    }
	}

      Logprint ("AbysS [PCAP] - Bringing up interface %s\n", dev);
      Logprint ("AbysS [Net] - Now entering passive OS detection mode\n");

      if (opts.bpf)
	Logprint ("AbysS [Net] - Using filter -> %s <-\n", filtr);

      pthread_create (&passive_thread, NULL, checkp, NULL);
      pthread_join (passive_thread, NULL);
    }

  if (opts.pingnet)
    {

      if ((ret = get_mask (opts.host)) == -1)
	{
	  printf ("AbysS [Error] - Incorrect netmask\n");
	  usage ();
	}

      if ((ret != 8) && (ret != 16) && (ret != 24))
	{
	  Logprint ("\nAbysS [Net] - Incorrect netmask\n");
	  usleep (250);
	  usage ();
	}


      if ((err = init_pcap_ ("icmp")) < 0)
	{
	  Logprint ("AbysS [Error] - Error opening up device %s\n", dev);
	  exit (-1);
	}

      if (pthread_create (&icmp_sniff_thread, NULL, icmp_handler, NULL))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));


      if (pthread_create (&tid, NULL, icmp_scan, NULL))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));

      pthread_mutex_lock (&finlock);

      while (finished == 0)

	pthread_cond_wait (&fincond, &finlock);
      if (global_thread_flag & DONE)

	pthread_join (tid, NULL);

      pthread_mutex_unlock (&finlock);


    }

  if (opts.synscan)
    {

      if ((opts.ip = resolve_ip (opts.host, (struct hostent *) &hptr)) == -1)
	{
	  printf ("AbysS [Net] - Could not resolve single host %s\n",
		  opts.host);
	  exit (0);
	}
      Logprint ("AbysS [Net] - %s has the ip address of %s\n",
		opts.host, inet_ntoa ((struct in_addr *) opts.ip));


      if (opts.verbose)
	{
	  i = 0;
	  hptr = hostent_ (opts.host);
	  for (ptr = hptr->h_aliases; *ptr != NULL; ptr++)

	    Logprint ("AbysS [Net] - Alias %d for %s : %s\n", ++i, opts.host,
		      *ptr);

	}

      if ((err = init_pcap_ ("tcp and src host %s", inet_ntoa (opts.ip))) < 0)
	{
	  Logprint ("AbysS [Error] - Error opening up device %s\n", dev);
	  exit (-1);
	}

      host[0].ip = ntohl (opts.ip);

      if (pthread_create (&syn_sniff_thread, NULL, syn_handler, NULL))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));


      if (pthread_create (&tid, NULL, scan, (void *) (host + 0)))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));


      pthread_mutex_lock (&finlock);
      while (finished == 0)
	pthread_cond_wait (&fincond, &finlock);

      if (global_thread_flag & DONE)
	{
	  pthread_join (tid, NULL);
	}
      pthread_mutex_unlock (&finlock);

      /* Give reasonable time for our SYN packet handler to procesess */
      /* the packets, and then kill the callback handler entirely */

      sleep (SYN_WAIT);
      Logprint ("AbysS [Net] - (UN-STEALTH) SYN Scan completed...\n");
      Logprint ("[Killing Callback handler now]\n");

      pthread_kill (syn_sniff_thread, SIGKILL);
      pthread_join (syn_sniff_thread, NULL);
    }

  if (opts.singlehost)
    {

      if ((opts.ip = resolve_ip (opts.host, (struct hostent *) &hptr)) == -1)
	{
	  printf ("AbysS [Net] - Could not resolve single host %s\n",
		  opts.host);
	  exit (0);
	}
      Logprint ("AbysS [Net] - %s has the ip address of %s\n",
		opts.host, inet_ntoa ((struct in_addr *) opts.ip));

      /* get aliases */
      if (opts.verbose)
	{
	  i = 0;
	  hptr = hostent_ (opts.host);
	  for (ptr = hptr->h_aliases; *ptr != NULL; ptr++)

	    Logprint ("AbysS [Net] - Alias %d for %s : %s\n", ++i, opts.host,
		      *ptr);
	}

      /* since we are only doing singlehost */
      /* setup on filter based on that ip   */

      if ((err = init_pcap_ ("tcp and src host %s", inet_ntoa (opts.ip))) < 0)
	{
	  Logprint ("AbysS [Error] - Error opening up device %s\n", dev);
	  exit (-1);
	}
      else
	Logprint ("AbysS [PCAP] - Bringing up interface %s\n", dev);

      host[0].ip = ntohl (opts.ip);

      /* Create sniff thread - note: checkp is a dummy function */
      /* After it is spawned by a thread, _checkp_ is called */
      /* as a call back handler by pcap_loop() */

      if (pthread_create (&sniff_thread, NULL, checkp, NULL))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));

      if (pthread_create (&tid, NULL, scan, (void *) (host + 0)))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));


      /* Wait until scan() sends us a signal saying finished is > 0 */

      pthread_mutex_lock (&finlock);
      while (finished == 0)
	pthread_cond_wait (&fincond, &finlock);

      if (global_thread_flag & DONE)
	{
	  pthread_join (sniff_thread, NULL);
	  pthread_join (tid, NULL);
	}
      pthread_mutex_unlock (&finlock);
    }

/* If subnet scanning is on, get_mask takes care of resolving the host */
/* And converting the ip to NBO, where it parses the binary value into */
/* Two seperate ip's representing the first and last ip spanning over */
/* the subnet class */

  if (opts.subscan)
    {

      if ((ret = get_mask (opts.host)) == -1)
	{
	  printf ("AbysS [Error] - Incorrect netmask\n");
	  usage ();
	}

      if ((ret != 8) && (ret != 16) && (ret != 24))
	{
	  Logprint ("\nAbysS [Net] - Incorrect netmask\n");
	  usleep (250);
	  usage ();
	}

      if ((err = init_pcap_ ("tcp")) < 0)
	{
	  Logprint ("AbysS [Error] - Opening up device %s\n", dev);
	  exit (-1);
	}

      if (pthread_create (&sniff_thread, NULL, checkp, NULL))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));

      /* jesus christ, inet_ntoa is non reentrant */
      /* use inet_ntop in the future */

      switch (ret)
	{
	case 8:
	  Logprint ("AbysS [Net] - %s -> Class A Netmask\n", opts.host);
	  Logprint ("AbysS [Net] - Start ip: %s\n",
		    inet_ntoa (addr.start_dest));
	  Logprint ("AbysS [Net] - Final ip: %s\n",
		    inet_ntoa (addr.final_dest));
	  opts.class = 0xFFFFFF;
	  sleep (1);
	  break;
	case 16:
	  Logprint ("AbysS [Net] - %s -> Class B Netmask\n", opts.host);
	  Logprint ("AbysS [Net] - Start ip: %s\n",
		    inet_ntoa (addr.start_dest));
	  Logprint ("AbysS [Net] - Final ip: %s\n",
		    inet_ntoa (addr.final_dest));
	  opts.class = 0xFFFF;
	  sleep (1);
	  break;
	case 24:
	  Logprint ("AbysS [Net] - %s -> Class C Netmask\n", opts.host);
	  Logprint ("AbysS [Net] - Start ip: %s\n",
		    inet_ntoa (addr.start_dest));
	  Logprint ("AbysS [Net] - Final ip: %s\n",
		    inet_ntoa (addr.final_dest));
	  opts.class = 0xFF;
	  sleep (1);
	  break;
	}

      /* Get maximum number of threads */

      nconn = 0;
      if (opts.threads)
	maxconn = opts.threads;
      else
	maxconn = 5;


      nconn = 0;
      for (net = ntohl (addr.start_dest); net < ntohl (addr.final_dest);
	   net++)
	{
	  if ((net & 0xFF) == 0)
	    net++;
	  if ((net & 0xFF) == 255)
	    net++;

	  /* We only cycle through max connections one time */
	  /* So we only create the first set of scan connections threads here */
	  /* After that, nconn will be maxed, and we will use thread condition signals */
	  /* Down below to check for finished threads and then create more */

	  while (nconn < maxconn)
	    {
	      host[nconn].ip = ntohl (net);
	      pthread_create (&host[nconn].tid, NULL, scan,
			      (void *) (host + nconn));
	      THREAD_WAIT;
	      nconn++;
	      net++;
	    }
	  /* Create a mutex, and condition, waiting for a thread to signify finished */
	  /* Is larger than one. After thread sends a signal saying finished is > 0 */
	  /* We cycle through the threads looking to see which it is thats done */
	  /* Get the next ip, and create a new thread */

	  pthread_mutex_lock (&finlock);
	  while (finished == 0)
	    pthread_cond_wait (&fincond, &finlock);

	  for (i = 0; i < opts.threads; i++)
	    {
	      if (host[i].socket.flag & DONE)
		{
		  pthread_join (host[i].tid, NULL);
		  host[i].ip = ntohl (net++);
		  host[i].socket.flag = 0;
		  THREAD_WAIT;
		  pthread_create (&host[i].tid, NULL, scan,
				  (void *) (host + i));
		  finished = 0;
		}
	    }
	  pthread_mutex_unlock (&finlock);
	}
      for (i = 0; i < maxconn; i++)
	pthread_join (host[i].tid, NULL);

    }
  nconn = 0;
  if (opts.threads)
    maxconn = opts.threads;
  if (opts.fromlist)
    {

      if ((fd = fopen (opts.hostlist, "r")) == NULL)
	{
	  Logprint ("AbysS [Error] - Could not open\n", opts.hostlist);
	  exit (0);
	}

      if ((err = init_pcap_ ("tcp")) < 0)
	{
	  Logprint ("AbysS [Error] - Opening up device %s\n", dev);
	  exit (-1);
	}

      if (pthread_create (&sniff_thread, NULL, checkp, NULL))
	fprintf (stderr, "AbysS [Error] - Could not create thread %s\n",
		 strerror (errno));

      i = 0;
      while (nconn < maxconn)
	{
	  if (fgets (buf, MAX_STRING, fd) != NULL)
	    {
	      if ((p2 = strchr (buf, '\n')) != NULL)
		*(p2) = '\0';

	      THREAD_WAIT;
	      host[nconn].name = strdup (buf);
	      pthread_create (&host[nconn].tid, NULL, scan,
			      (void *) (host + nconn));
	      nconn++;
	    }
	}
      fseek (fd, 0, SEEK_CUR);
      memset (buf, 0, MAX_STRING);
      while ((fgets (buf, MAX_STRING, fd)) != NULL)
	{
	  if ((p2 = strchr (buf, '\n')) != NULL)
	    *(p2) = '\0';

	  /* Same thing again with the mutex, and waiting for the thread condition signal */
	  /* Things are only changed in the sense that we are reading from a host list */

	  pthread_mutex_lock (&finlock);
	  while (finished == 0)
	    pthread_cond_wait (&fincond, &finlock);

	  for (i = 0; i < opts.threads; i++)
	    {
	      if (host[i].socket.flag & DONE)
		{
		  pthread_join (host[i].tid, NULL);
		  host[i].name = strdup (buf);
		  host[i].socket.flag = 0;
		  THREAD_WAIT;
		  pthread_create (&host[i].tid, NULL, scan,
				  (void *) (host + i));
		  finished = 0;
		}
	    }
	  pthread_mutex_unlock (&finlock);
	}

    }

final:

  if (opts.subscan)
    {
      for (i = 0; i < maxconn; i++)
	free (host[i].name);
      free (dev);
    }

  if (opts.subscan)
    {
      for (i = 0; i < maxconn; i++)
	pthread_join (host[i].tid, NULL);
      pthread_join (sniff_thread, NULL);
    }

  if (opts.fromlist)
    {
      for (i = 0; i < maxconn; i++)
	pthread_join (host[i].tid, NULL);
      pthread_join (sniff_thread, NULL);
    }


  if (opts.logging)
    free (opts.logfile);
  if (opts.fromlist)
    free (opts.hostlist);
  if (opts.singlehost)
    free (opts.host);

  pcap_done_ ();
  time (&clock);
  tm = gmtime (&clock);
  setlocale (LC_TIME, "C");
  strftime (time_buf, sizeof (time_buf), "%A, %B %e, %Y, %H:%M:%S", tm);
  Logprint ("AbysS [Report] - Scan finished -> [%s]\n", time_buf);

  exit (0);
}
